subroutine deposit_cic(nx, ny, x_min, y_min, dx, dy, &
    x, y, q, n_particles, rho)
    use omp_lib
    implicit none

    integer, intent(in) :: nx, ny       ! Number of grid points in x and y directions
    real(8), intent(in) :: x_min, y_min ! Coordinates of a center of the most lower left grid cell
    real(8), intent(in) :: dx, dy       ! Grid spacing
    real(8), intent(in) :: q ! Charge per macroparticle
    integer, intent(in) :: n_particles  ! Number of particles
    real(8), intent(in) :: x(n_particles), y(n_particles) ! Particle positions

    real(8), intent(inout) :: rho(nx, ny)  ! Charge density array

    integer :: pid, i, j
    real(8) :: wx, wy
    !$omp parallel do private(i, i_grid, j_grid, wx, wy)
    do pid = 1, n_particles

        i = int((x(pid) - x_min) / dx)
        j = int((y(pid) - y_min) / dy)

        wx = ((x(pid) - x_min) / dx) - dble(i)
        wy = ((y(pid) - y_min) / dy) - dble(j)

        if (i >= 0 .and. i < nx-1 .and. j >= 0 .and. j < ny-1) then
            !$omp atomic
            rho(i+1, j+1) = rho(i+1, j+1) + q * (1.0d0 - wx) * (1.0d0 - wy)
            !$omp atomic
            rho(i+2, j+1) = rho(i+2, j+1) + q * wx * (1.0d0 - wy)
            !$omp atomic
            rho(i+1, j+2) = rho(i+1, j+2) + q * (1.0d0 - wx) * wy
            !$omp atomic
            rho(i+2, j+2) = rho(i+2, j+2) + q * wx * wy
        end if
    end do
    !$omp end parallel do

end subroutine deposit_cic

subroutine interpolate_electric_field(nx, ny, n_particles, x_min, y_min, dx, dy, &
    x_particles, y_particles, Ex_grid, Ey_grid, &
    Ex_interp, Ey_interp)
    use omp_lib
    implicit none

    ! Input parameters
    integer, intent(in) :: nx, ny, n_particles  ! Number of grid points and particles
    real(8), intent(in) :: x_min, y_min  ! Domain bounds
    real(8), intent(in) :: dx, dy       ! Grid spacing
    real(8), intent(in) :: x_particles(n_particles), y_particles(n_particles)  ! Particle positions
    real(8), intent(in) :: Ex_grid(nx, ny), Ey_grid(nx, ny)  ! Grid electric fields

    ! Output parameters
    real(8), intent(out) :: Ex_interp(n_particles), Ey_interp(n_particles)  ! Interpolated fields

    ! Local variables
    integer :: i, j, pid
    real(8) :: wx, wy, ex, ey

    ! Loop over each particle to interpolate electric fields
    !$omp parallel do private(i, j, px, x_rel, y_rel, wx, wy, ex, ey)
    do pid = 1, n_particles
        ! Find the indices of the lower-left grid point
        i = int((x_particles(pid) - x_min) / dx)
        j = int((y_particles(pid) - y_min) / dy)

        ! Relative positions within the cell
        wx = ((x_particles(pid) - x_min) / dx) - dble(i)
        wy = ((y_particles(pid) - y_min) / dy) - dble(j)

        ! Check bounds and perform bilinear interpolation
        if (i >= 0 .and. i < nx-1 .and. j >= 0 .and. j < ny-1) then
            ex = (1.0d0 - wx) * (1.0d0 - wy) * Ex_grid(i+1, j+1) + &
            wx * (1.0d0 - wy) * Ex_grid(i+2, j+1) + &
            (1.0d0 - wx) * wy * Ex_grid(i+1, j+2) + &
            wx * wy * Ex_grid(i+2, j+2)

            ey = (1.0d0 - wx) * (1.0d0 - wy) * Ey_grid(i+1, j+1) + &
            wx * (1.0d0 - wy) * Ey_grid(i+2, j+1) + &
            (1.0d0 - wx) * wy * Ey_grid(i+1, j+2) + &
            wx * wy * Ey_grid(i+2, j+2)

            Ex_interp(pid) = ex
            Ey_interp(pid) = ey
        else
            Ex_interp(pid) = 0.0d0
            Ey_interp(pid) = 0.0d0
        end if
    end do
    !$omp end parallel do

end subroutine interpolate_electric_field
