import numpy as np
from scipy.constants import epsilon_0

import mbtrack2.pic.grid2d as grid2d


class PICSolver2D:
    """
        A 2D Particle-In-Cell (PIC) solver for solving the Poisson equation using
        Integrated Green's Function (IGF) methods. This class handles
        charge deposition, potential calculation, and electric field computation
        on a 2D grid.

        Parameters
        ----------
        grid_size : tuple of float
            A tuple (x_min, x_max, y_min, y_max) defining the physical boundaries
            of the grid.
        cell_size : tuple of float
            A tuple (dx, dy) defining the size of each cell in the grid.

        Attributes
        ----------
        x_min, x_max, y_min, y_max : float
            The physical boundaries of the grid.
        dx, dy : float
            The size of each cell in the grid.
        nx, ny : int
            The number of grid points in the x and y directions.
        rho : numpy.ndarray
            The charge density array.
        phi : numpy.ndarray
            The electric potential array.
        k_squared : numpy.ndarray
            The squared wavenumber array used in FFT-based Poisson solver.
        green_hat : numpy.ndarray
            The Fourier transform of the expanded Green's function.

        Methods
        -------
        update_grid_igf(grid_size, cell_size)
            Updates the grid parameters and precomputes the Green's function.
        deposit_cic(x, y, q)
            Deposits charge onto the grid using the Cloud-In-Cell (CIC) method.
        reset_deposit()
            Resets the charge density array.
        poisson_solver_2d_fft()
            Solves the 2D Poisson equation using FFT.
        poisson_solver_2d_igf()
            Solves the 2D Poisson equation using the Integrated Green's Function.
        calculate_electric_field()
            Calculates the electric field from the potential.
        interpolate_electric_field(x_particles, y_particles)
            Interpolates the electric field at particle positions.
    """

    def __init__(self, grid_size, cell_size):
        """
        Initializes the PICSolver2D with grid size and cell size.

        Parameters
        ----------
        grid_size : tuple of float
            A tuple (x_min, x_max, y_min, y_max) defining the physical boundaries
            of the grid.
        cell_size : tuple of float
            A tuple (dx, dy) defining the size of each cell in the grid.
        """
        self.update_grid_igf(grid_size, cell_size)
        kx = np.fft.fftfreq(self.nx, d=self.dx) * 2 * np.pi
        ky = np.fft.fftfreq(self.ny, d=self.dy) * 2 * np.pi
        kx, ky = np.meshgrid(kx, ky, indexing='ij')
        self.k_squared = kx*kx + ky*ky

        self.k_squared[
            0, 0] = 1.0  # Prevents division by zero at (kx, ky) = (0, 0)

    def update_grid_igf(self, grid_size, cell_size):
        """
        Updates the grid parameters and precomputes the Green's function.

        Parameters
        ----------
        grid_size : tuple of float
            A tuple (x_min, x_max, y_min, y_max) defining the physical boundaries
            of the grid.
        cell_size : tuple of float
            A tuple (dx, dy) defining the size of each cell in the grid.
        """
        self.x_min, self.x_max, self.y_min, self.y_max = grid_size
        self.dx, self.dy = cell_size

        self.nx, self.ny = map(
            lambda delta, d: (delta/d + 1).astype(np.int64),
            [self.x_max - self.x_min, self.y_max - self.y_min],
            [self.dx, self.dy])

        self.rho = np.zeros((self.nx, self.ny), order='F', dtype=np.float64)
        self.phi = np.zeros((self.nx, self.ny), dtype=np.float64)
        green_function = np.zeros((2 * self.nx, 2 * self.ny))

        # Grid for green's function should have an origin of (0, 0) for the
        # convolution to work
        X, Y = np.meshgrid(np.arange(self.nx + 2) * self.dx - self.dx / 2,
                           np.arange(self.ny + 2) * self.dy - self.dy / 2,
                           indexing='ij')
        green_function0 = self._compute_integrated_green_function(X, Y)

        # Expanding the domain of Green's function by mirroring it while
        # maintaining periodicity of double the grid size. The Green function
        # indices in each dimension should be 0...N-1, N, N-1, ..., 1.
        # This is the Hockney-Eastwood method (add citation).
        green_function0 = green_function0[
            1:, 1:] + green_function0[:-1, :-1] - green_function0[
                1:, :-1] - green_function0[:-1, 1:]
        green_function[:self.nx + 1, :self.ny + 1] = green_function0
        green_function[self.nx + 1:, :self.ny + 1] = green_function0[self.nx -
                                                                     1:0:-1, :]
        green_function[:self.nx + 1,
                       self.ny + 1:] = green_function0[:, self.ny - 1:0:-1]
        green_function[self.nx + 1:,
                       self.ny + 1:] = green_function0[self.nx - 1:0:-1,
                                                       self.ny - 1:0:-1]

        self.green_hat = np.fft.fft2(green_function,
                                     s=(2 * self.nx, 2 * self.ny))

    def deposit_cic(self, x, y, q):
        """
        Deposits charge onto the grid using the Cloud-In-Cell (CIC) method.
        Implemented in fortran pic2d.f90.

        Parameters
        ----------
        x : numpy.ndarray
            The x-coordinates of the particles.
        y : numpy.ndarray
            The y-coordinates of the particles.
        q : numpy.ndarray
            The charges of the particles.
        """
        grid2d.deposit_cic(self.x_min, self.y_min, self.dx, self.dy, x, y, q,
                           self.rho)
        total_charge = self.rho.sum()
        if total_charge != 0:
            self.rho /= (self.dx * self.dy * total_charge)

    def reset_deposit(self):
        """
        Resets the charge density array to zero.
        """
        self.rho = np.zeros((self.nx, self.ny), order='F', dtype=np.float64)

    def poisson_solver_2d_fft(self):
        """
        Solves the 2D Poisson equation with periodic (?) boundary conditions using FFT.
        """

        rho_hat = np.fft.fft2(self.rho, s=(self.nx, self.ny))

        phi_hat = rho_hat / self.k_squared / epsilon_0

        phi_hat[0, 0] = 0.0

        self.phi = np.fft.ifft2(phi_hat, s=(self.nx, self.ny)).real

    def _compute_integrated_green_function(self, x, y):
        """
        Computes the integrated Green's function for the given coordinates.

        Parameters
        ----------
        x : numpy.ndarray
            The x-coordinates.
        y : numpy.ndarray
            The y-coordinates.

        Returns
        -------
        numpy.ndarray
            The integrated Green's function values.
        """
        return 1 / (4 * np.pi * epsilon_0) * (
            3*x*y - x * x * np.arctan(y / x) - y * y * np.arctan(x / y) -
            x * y * np.log(x*x + y*y))

    def poisson_solver_2d_igf(self):
        """
        Solves the 2D Poisson equation using the Integrated Green's Function
        approach with open boundary conditions.
        """
        rho = np.zeros((2 * self.nx, 2 * self.ny))

        rho[:self.nx, :self.ny] = self.rho
        rho_hat = np.fft.fft2(rho, s=(2 * self.nx, 2 * self.ny))

        self.phi = np.fft.ifft2(self.green_hat * rho_hat,
                                s=(2 * self.nx,
                                   2 * self.ny)).real[:self.nx, :self.ny]

    def _calculate_electric_field(self):
        """
        Calculates the electric field from the potential.

        Returns
        -------
        E_x, E_y : numpy.ndarray
            The x and y components of the electric field.
        """
        E_x, E_y = np.gradient(-self.phi, self.dx, self.dy)
        return E_x, E_y

    def interpolate_electric_field(self, x_particles, y_particles):
        """
        Interpolates the electric field at particle positions. Interpolation is
        implemented in grid2d.f90.

        Parameters
        ----------
        x_particles : numpy.ndarray
            The x-coordinates of the particles.
        y_particles : numpy.ndarray
            The y-coordinates of the particles.

        Returns
        -------
        E_x_interp, E_y_interp : numpy.ndarray
            The interpolated x and y components of the electric field at particle positions.
        """
        E_x, E_y = self._calculate_electric_field()
        E_x_interp, E_y_interp = grid2d.interpolate_electric_field(
            self.x_min, self.y_min, self.dx, self.dy, x_particles, y_particles,
            E_x, E_y)
        return E_x_interp, E_y_interp
